"""Aggregate aspects into semantic clusters ('topics')."""
import logging
from collections import Counter

import networkx as nx
import numpy as np
import pandas as pd
from networkx.algorithms import approximation
from nltk import tokenize
from nltk.corpus import wordnet as wn
from scipy.spatial import distance

from extra_model._disambiguate import match

logger = logging.getLogger(__name__)


def path_to_graph(hypernym_list, initialnoun):
    """Make a hypernym chain into a graph.

    :param hypernym_list: list of hypernyms for a word as obtained from wordnet
    :type hypernym_list: [str]
    :param initialnoun: the initial noun (we need this to mark it as leaf in the tree)
    :type initialnoun: str
    :return: the linear directed graph of the chain
    :rtype: :class:`networkx.DiGraph`
    """
    graph = nx.DiGraph()
    # mark the original word as 'seed' so we can track 'importance' later
    graph.add_node(initialnoun, seed=True)
    previous = initialnoun
    for hypernym in reversed(hypernym_list):
        # we'll take care of the distances later
        graph.add_edge(previous, hypernym.name(), similarity=1.0, distance=1.0)
        graph.nodes[hypernym.name()]["seed"] = False
        previous = hypernym.name()
    return graph


def get_nodevec(node, vectors):
    """Get the vector representation of a gloss a wordnet node.

    Used to evaluate similarity between rungs in the hypernym chain.
    :param node: the wornet node for which to compute the embedding
    :type node: str
    :param vectors: the vectorizer to use for the embedding
    :type vectors: :class:`extra_model._vectorizer.Vectorizer`
    :return: the embedding for the node
    :retype: :class:`numpy.array`
    """
    desc = tokenize.word_tokenize(wn.synset(node).definition())
    desc = [vectors.get_vector(word) for word in desc]
    desc = [word for word in desc if word is not None]
    desc = np.sum(desc, axis=0)
    desc = np.divide(desc, np.linalg.norm(desc))
    return desc


def iterate(transition_matrix, importance, original, alpha):
    """Find the stable importance vector by iterated multiplication with the distance matrix.

    This function does a simple iteration. The "jump-back" probability from the paper
    is implemented as a linear superposition of the new and original importance numbers.
    :param transition_matrix: The connectedness matrix of the graph, including similarity weights.
    :type transition_matrix: :class:`numpy.array`
    :param importance: Current importance vector
    :type importance: :class:`numpy.array`
    :param original: Original importance vector (i.e. aspect counts for leaf nodes, zero otherwise)
    :type original: :class:`numpy.array`
    :param alpha: jump-back probability
    :type alpha: float
    :return: the importance vector for the next step of the iteration
    :rtype: :class:`numpy.array`
    """
    importance = np.matmul(transition_matrix, importance)
    importance = np.multiply(importance, alpha)
    importance = np.add(
        importance, np.multiply(np.divide(original, np.sum(original)), 1.0 - alpha)
    )
    importance = np.divide(importance, np.sum(importance))
    return importance


def aggregate(aspects, aspect_counts, synsets_match, vectors):  # noqa: C901
    """Aggregate the aspects by building a tree from the hypernym chains.

    Using a page-rank type algorithm to assign importance to the nodes in the graph
    we only consider wordnet entries for this, not the actual aspects extracted from the texts.
    :param aspects: List of aspects to analyze
    :type aspects: [str]
    :param aspect_counts: Map with the aspects as keys and their counts as values
    :type aspect_counts: {str:int}
    :param synsets_match: List of synsets matched to each aspect
    :type synsets_match: [:class:`nltk.wordnet.Synset`]
    :param vectors: the vectorizer to use for the embedding
    :type vectors: :class:`extra_model._vectorizer.Vectorizer`
    :return: A list nodes and their importances
    :rtype: [(str,float)]
    """
    # count how many aspects are matched to a given wornet entry, so that we
    # can remove ambiguities from the graph

    synset_matchnumbers = {}
    for syns in synsets_match:
        if syns in synset_matchnumbers:
            synset_matchnumbers[syns] = synset_matchnumbers[syns] + 1
        else:
            synset_matchnumbers[syns] = 1

    # produce the tree from the hypernym chains
    # a linear graph for every hypernym chain starting from every original
    # aspect
    linear_graphs = []
    for noun, meaning in zip(aspects, synsets_match):
        if meaning is None:  # my life, my code
            continue
        hypernym_paths = meaning.hypernym_paths()
        # if only one hypernym path exists we don't have to select by
        # length/number of connected aspects
        if len(hypernym_paths) == 1:
            linear_graphs.append(path_to_graph(hypernym_paths[0], noun))
            continue
        path_match_nums = [
            np.sum(
                [
                    synset_matchnumbers[syns] if syns in synset_matchnumbers else 0
                    for syns in path
                ]
            )
            for path in hypernym_paths
        ]
        max_match = max(path_match_nums)
        # we need to do the forloop here, because it's possible to have two
        # chains with equal numbers of attached aspects
        for n in range(len(hypernym_paths)):
            if path_match_nums[n] == max_match:
                linear_graphs.append(path_to_graph(hypernym_paths[n], noun))

    # create the full graph as union of all the hypernym chains and lable
    # leaf-nodes
    full_tree = nx.compose_all(linear_graphs)
    # add similarity scores
    for from_node, to_node in full_tree.edges():
        # similarity is only useful within wordnet
        if not full_tree.nodes[from_node]["seed"]:
            full_tree.edges[from_node, to_node]["distance"] = distance.cosine(
                get_nodevec(from_node, vectors), get_nodevec(to_node, vectors)
            )
            full_tree.edges[from_node, to_node]["similarity"] = (
                1.0 - full_tree.edges[from_node, to_node]["distance"]
            )

    # Remove loops with steiner approximation
    seeds = [
        node for node, node_data in full_tree.nodes(data=True) if node_data["seed"]
    ]
    pruned_tree = approximation.steiner_tree(
        full_tree.to_undirected(), seeds, weight="distance"
    )
    pruned_nodes = set(pruned_tree.nodes())
    nodes_to_remove = set(full_tree.nodes())
    nodes_to_remove.difference_update(pruned_nodes)
    full_tree.remove_nodes_from(nodes_to_remove)
    wordnet_nodes = [
        node for node, node_data in full_tree.nodes(data=True) if not node_data["seed"]
    ]

    # produce the similarity matrix
    transition_matrix = np.zeros((len(wordnet_nodes), len(wordnet_nodes)))
    # diagonal needs to be unity
    for n_node in range(len(wordnet_nodes)):
        transition_matrix[n_node][n_node] = 1.0
    wordnet_nodes_indices = dict(
        zip(wordnet_nodes, [i for i in range(len(wordnet_nodes))])
    )
    for from_node, to_node, similarity in full_tree.edges(data="similarity"):
        if not full_tree.nodes[from_node]["seed"]:  # skip the seed nodes for this
            transition_matrix[wordnet_nodes_indices[to_node]][
                wordnet_nodes_indices[from_node]
            ] = similarity

    # start with the an innitial importance weight proportional to the number
    # of mentions of linked aspects in the reviews
    importance_vec = []
    for node in wordnet_nodes:
        imp = 0
        for neighbor in full_tree.predecessors(node):
            if full_tree.nodes[neighbor]["seed"]:
                imp = imp + aspect_counts[neighbor]
        importance_vec.append(imp)

    importance = np.array(importance_vec)
    importance = np.divide(importance, np.sum(importance))
    importance_original = importance.copy()

    # find the fixed point of the page-rank-algorithm
    for i in range(100):
        importance = iterate(
            transition_matrix, importance, importance_original, 0.5
        )  # XXX make alpha configurable

    # sort the resulting topics by importnace
    topics = list(zip(wordnet_nodes, importance))
    topics = sorted(topics, key=lambda pair: pair[1], reverse=True)
    logger.debug(topics)
    return topics, full_tree  # we still need the tree for filtering


def traverse_tree(  # noqa: C901
    node_list, associated_aspects, aspect_counts, full_tree, weighted, direction
):
    """Find all hypernyms/hyponyms in the tree to a given node.

    Aggregate the number of associated mentions in the original texts, optionally
    weighted by term-similarity.
    :param nodelist: List of nodes from which to gather the subsidiary terms and their initial mentions
    :type nodelist: [(str,int)]
    :param associated_aspects:  intermediate map of associated terms, used in recursion, start with an empty map
    :type associated_aspects: {str:float}
    :param aspect_counts: List of the pairs of aspects and their counts
    :type aspect_counts: [(str,int)]
    :param full_tree: the graph which is being traversed
    :type full_tree: :class:`networkx.DiGraph`
    :param weighted: True->similarity weights are used in aggregation. False->weights are assumed to be 1
    :type weighted: Bool
    :param direction: In which direction to traverse the tree "up" or "down"
    :type direction: str
    :return: A map of the resulting terms associated to the initial node and their distance-weighted number of mentions
    :rtype: {str:float}
    """
    new_nodes = []
    for node, weight in node_list:
        for daughter in full_tree.predecessors(node):
            # we need the directed unpruned graph here to avoid descent from
            # hypernyms
            if full_tree.nodes[daughter]["seed"] and daughter not in associated_aspects:
                associated_aspects[daughter] = aspect_counts[daughter] * weight
        maybe_daugthers = None
        if direction == "up":
            maybe_daugthers = full_tree.neighbors(node)
        if direction == "down":
            maybe_daugthers = full_tree.predecessors(node)
        for daughter in maybe_daugthers:
            if full_tree.nodes[daughter]["seed"]:  # we already took care of these
                continue
            elif weighted:  # go further up/down the tree
                new_nodes.append(
                    (
                        daughter,
                        weight
                        * (
                            full_tree.edges[node, daughter]["similarity"]
                            if direction == "up"
                            else full_tree.edges[daughter, node]["similarity"]
                        ),
                    )
                )
            else:
                new_nodes.append((daughter, 1))
    if len(new_nodes) == 0:
        return associated_aspects
    else:
        return traverse_tree(
            new_nodes, associated_aspects, aspect_counts, full_tree, weighted, direction
        )


def collect_topic_info(filtered_topics, removed_topics, aspect_counts, full_tree):
    """Gather various bits of information into a single DataFrame.

    For each topic we store the importance, the list of associated raw text terms and their numbers.
    :param filtered_topics: List of topics remaining after filtering out low-iimportance subsidiary topics
    :type filtered_topics: [str]
    :param removed_topics: Map with the removed topics keyed to the selected topics to which they are subsidiary
    :param removed_topics: {str:[str]}
    :param aspect_counts: List of the pairs of aspects and their counts
    :type aspect_counts: [(str,int)]
    :param full_tree: the graph which is being traversed
    :type full_tree: :class:`networkx.DiGraph`
    :return: the aggregated dataframe of topics
    :rtype: :class:`pandas.DataFrame`
    """
    row_vec = []
    for topic in filtered_topics:
        topic_noun = topic[0]
        row = {
            "topic": topic_noun,
            "importance": topic[1],
            "rawterms": [],
            "rawnums": [],
            "weightedterms": [],
            "weights": [],
        }
        logger.debug("Common terms for aspect {0!s} (by raw number)".format(topic_noun))
        raw_aspects = {}
        subsidiaries = removed_topics[topic_noun]
        subsidiaries.append(topic_noun)
        logger.debug(subsidiaries)
        for subtopic in subsidiaries:
            for aspect, count in aspect_counts.items():
                if full_tree.has_edge(aspect, subtopic):
                    raw_aspects[aspect] = count
        if len(raw_aspects) == 0:
            logger.debug("No attached terms, skipping")
            continue
        ordered_aspects = sorted(
            raw_aspects.items(), key=lambda pair: pair[1], reverse=True
        )
        row["rawterms"] = list(list(zip(*ordered_aspects))[0])
        row["rawnums"] = list(list(zip(*ordered_aspects))[1])
        for j in range(min(20, len(ordered_aspects))):  # print some exampels
            logger.debug("{} {}".format(ordered_aspects[j][0], ordered_aspects[j][1]))
        logger.debug(
            "Common terms for aspect {0!s} (distance weighted)".format(topic_noun)
        )
        weighted_aspects = traverse_tree(
            [(topic_noun, 1)],
            {},
            aspect_counts,
            full_tree,
            weighted=True,
            direction="down",
        )
        ordered_aspects_weighted = sorted(
            weighted_aspects.items(), key=lambda pair: pair[1], reverse=True
        )
        row["weightedterms"] = list(list(zip(*ordered_aspects_weighted))[0])
        row["weights"] = list(list(zip(*ordered_aspects_weighted))[1])
        for j in range(min(20, len(ordered_aspects_weighted))):
            logger.debug(
                "{} {}".format(
                    ordered_aspects_weighted[j][0], ordered_aspects_weighted[j][1]
                )
            )
        row_vec.append(row)

    dataframe_topics = pd.DataFrame(
        row_vec,
        columns=[
            "topic",
            "importance",
            "rawterms",
            "rawnums",
            "weightedterms",
            "weights",
        ],
    )
    return dataframe_topics


def has_connection(term, prior, full_tree):
    """Check if two terms are connected within the directed hyopernym graph.

    :param term: first node to test
    :type term: str
    :param prior: second node to test
    :type prior: str
    :param full_tree: the graph which is being traversed
    :type full_tree: :class:`networkx.DiGraph`
    :return: Do the two nodes have a connection in the directed graph?
    :rtype: Bool
    """
    if term in nx.descendants(full_tree, prior) or prior in nx.descendants(
        full_tree, term
    ):
        return True
    return False


def filter_aggregates(topics, tree):
    """Filter the importance-sorted list, so that each remaining topic is the sole member of its hypernym chain.

    :param topics: List of all topics in the graph
    :type topics: [str]
    :param tree: the graph which is being traversed
    :type tree: :class:`networkx.DiGraph`
    :return: a tuple of 1) the list of surviving topics, 2) a map of the culled topics keyed to the topics they are subsidiary to
    :rtype: ([str],{str:[str]})
    """
    filtered_topics = []
    removed_topics = {}
    for term in topics:
        is_new = True
        colliding_prior = None
        for prior in filtered_topics:
            if has_connection(term[0], prior[0], tree):
                is_new = False
                colliding_prior = prior[0]
                break
        if is_new:
            filtered_topics.append(term)
            removed_topics[term[0]] = []
        else:
            removed_topics[colliding_prior].append(term[0])
            logger.debug("Dropping aspect: {0!s}".format(term))
    return filtered_topics, removed_topics


def get_topics(dataframe_aspects, vectors):
    """Generate the semantically clustered topics from the raw aspects.

    :param dataframe_aspects: the collection of nouns to be aggregated into topics
    :type dataframe_aspects: :class:`pandas.DataFrame`
    :param vectors: provides embeddings for context clustering and wordsense disammbguation
    :type vectors: :class:`extra_model._vectorizer.Vectorizer`
    :return: The dataframe containing the topics and associated info
    :rtype: :class:`pandas.DataFrame`
    """
    # for most of the processing we don't really need the specific aspect
    # instances, but just a dict with the aspects and numbers of appearance
    aspect_counts = Counter(dataframe_aspects["aspect"])
    dataframe_aspects.loc[:, "aspect_count"] = dataframe_aspects["aspect"].apply(
        lambda aspect: aspect_counts[aspect]
    )
    logger.debug(aspect_counts.most_common())

    # match the aspects to dictionary terms, filtering aspects that can't be
    # mathched
    aspects, synsets_match = match(aspect_counts, vectors)
    disambiguation_dict = dict(zip(aspects, synsets_match))
    dataframe_aspects.loc[:, "wordnet_node"] = dataframe_aspects["aspect"].apply(
        lambda aspect: disambiguation_dict[aspect].name()
        if (aspect in disambiguation_dict and disambiguation_dict[aspect] is not None)
        else None
    )

    # build the hypernym graph
    topics, full_tree = aggregate(aspects, aspect_counts, synsets_match, vectors)

    # filter the importance ranked topics, so that only one topic within a
    # given hypernym chain remains
    filtered_topics, removed_topics = filter_aggregates(topics, full_tree)

    logger.debug(filtered_topics)

    # gather information into a more structured format
    dataframe_topics = collect_topic_info(
        filtered_topics, removed_topics, aspect_counts, full_tree
    )

    return dataframe_topics



import os

import networkx as nx
import numpy as np
import pytest
from gensim.models import KeyedVectors
from nltk.corpus import wordnet as wn

from extra_model._topics import (
    aggregate,
    filter_aggregates,
    get_nodevec,
    has_connection,
    iterate,
    path_to_graph,
    traverse_tree,
)
from extra_model._vectorizer import Vectorizer


@pytest.fixture()
def minivec():
    # preprocess plain-text test embeddings to proper binary format for vectorizer
    glove_file = "tests/resources/test_topics.vec"
    prepro_file = "tests/resources/test_topics.prepro"
    model = KeyedVectors.load_word2vec_format(glove_file, binary=False, no_header=True)
    model.save(prepro_file)

    yield Vectorizer(prepro_file)

    # cleanup
    os.remove(prepro_file)


@pytest.fixture()
def vec():
    return Vectorizer("tests/resources/small_embeddings")


@pytest.fixture()
def simple_graph():
    # a simple example graph: two leaf nodes (corresponding to text instances), L1 and L2, connected to
    # the root of the tree R via two intermidate nodes I1 and I2 which represent higher-level wordnet synsets
    # the edges are weighted to test the similarity functionality
    graph = nx.DiGraph()
    graph.add_nodes_from(["L1", "I1", "L2", "I2", "R"])
    graph.add_edges_from([("L1", "I1"), ("L2", "I2"), ("I1", "R"), ("I2", "R")])
    graph.nodes["L1"]["seed"] = True
    graph.nodes["L2"]["seed"] = True
    graph.nodes["I1"]["seed"] = False
    graph.nodes["I2"]["seed"] = False
    graph.nodes["R"]["seed"] = False
    graph["L1"]["I1"]["similarity"] = 1.0
    graph["L2"]["I2"]["similarity"] = 1.0
    graph["I1"]["R"]["similarity"] = 0.5
    graph["I2"]["R"]["similarity"] = 0.5
    return graph


def test__path_to_graph():
    noun = "zombie"
    hypernym_list = wn.synsets(noun, pos=wn.NOUN)[0].hypernym_paths()[0]
    graph = path_to_graph(hypernym_list, noun)
    print(list(graph.successors(noun)))
    print(list(graph.successors("entity.n.01")))
    assert (
        # entity is the root
        len(list(graph.successors("entity.n.01"))) == 0
        # initial word is the leaf
        and len(nx.dfs_successors(graph, noun)) == graph.order() - 1
        # leaf has the seed-flag set
        and graph.nodes[noun]["seed"] is True
    )
    # all the other nodes should not be seeds
    for depth, nodes in nx.dfs_successors(graph, noun).items():
        for node in nodes:
            assert not graph.nodes[node]["seed"]
    # all intermediate nodes should have 2 connections, the leaf only one.
    for depth, nodes in nx.dfs_predecessors(graph, "entity.n.01").items():
        for node in nodes:
            if graph.nodes[node]["seed"] is True:
                assert graph.degree(node) == 1
            else:
                assert graph.degree(node) == 2
    print(graph.nodes)


def test__get_nodevec(minivec):
    assert (get_nodevec("entity.n.01", minivec) == [1.0, 0.0, 0.0, 0.0]).all()


def test__iterate():
    transition_matrix = np.array([[0.0, 1.0], [1.0, 0.0]])
    original = np.array([0.2, 0.8])
    importance = original.copy()
    result = iterate(transition_matrix, importance, original, 0.5)
    assert result[0] == pytest.approx(0.5)
    assert result[1] == pytest.approx(0.5)


def test__has_connection(simple_graph):
    assert (
        has_connection("L1", "R", simple_graph)
        and has_connection("R", "L1", simple_graph)
        and not has_connection("L1", "L2", simple_graph)
    )


def test__filter_aggregates(simple_graph):
    topics = [("L1", 0), ("I1", 0), ("I2", 0), ("L2", 0)]
    filtered, removed = filter_aggregates(topics, simple_graph)
    assert filtered == [("L1", 0), ("I2", 0)] and removed == {
        "I2": ["L2"],
        "L1": ["I1"],
    }


def test__aggregate(vec):
    res, tree = aggregate(
        ["chair", "Chair", "zombie", "unknown"],
        {"chair": 4, "Chair": 2, "zombie": 1, "unknown": 1},
        [
            wn.synset("chair.n.01"),
            wn.synset("chair.n.01"),
            wn.synset("zombi.n.01"),
            None,
        ],
        vec,
    )
    nodes, importances = zip(*res)
    assert (  # first entry should have highest importance because we put high count to it
        importances.index(max(importances)) == 0
    )


def test__traverse_tree__down_weighted(simple_graph):
    assert (
        traverse_tree(
            [("R", 1)],
            {},
            {"L1": 4, "L2": 1},
            simple_graph,
            weighted=True,
            direction="down",
        )
        == {"L1": 2, "L2": 0.5}
    )


def test__traverse_tree__down_unweighted(simple_graph):
    assert (
        traverse_tree(
            [("R", 1)],
            {},
            {"L1": 4, "L2": 1},
            simple_graph,
            weighted=False,
            direction="down",
        )
        == {"L1": 4, "L2": 1}
    )


def test__traverse_tree__up_weighted(simple_graph):
    assert traverse_tree(
        [("I1", 1)], {}, {"L1": 4, "L2": 1}, simple_graph, weighted=True, direction="up"
    ) == {"L1": 4}


def test__traverse_tree__up_unweighted(simple_graph):
    assert (
        traverse_tree(
            [("I1", 1)],
            {},
            {"L1": 4, "L2": 1},
            simple_graph,
            weighted=False,
            direction="up",
        )
        == {"L1": 4}
    )


@pytest.mark.skip(
    reason="This function is hard to test in isolation and is tested as part of the integration test"
)
def test_aspects__get_topics():
    pass


@pytest.mark.skip(
    reason="This function is hard to test in isolation and is tested as part of the integration test"
)
def test_aspects__collect_topic_info():
    pass



import pandas as pd
import pytest

from extra_model._filter import filter


@pytest.fixture()
def dataset():
    input_df = pd.DataFrame(
        [
            {"Comments": "This is an example of English comment", "id": "good"},
            {
                "Comments": "Dies ist ein Beispiel für einen deutschen Kommentar",
                "id": "german",
            },
            {"Comments": "Short", "id": "short"},
            {
                "Comments": "This is a comment with \n non-printable sequence",
                "id": "nonprintable",
            },
        ]
    )

    return input_df


@pytest.fixture()
def cleaned_dataset(dataset):
    cleaned = filter(dataset)
    return cleaned


def test__dropping_short_comment(cleaned_dataset):
    ids = cleaned_dataset["id"].tolist()
    assert "short" not in ids


def test__dropping_non_english_comment(cleaned_dataset):
    ids = cleaned_dataset["id"].tolist()
    assert "german" not in ids


def test__no_unreadable_characters(cleaned_dataset):
    nonprintable_id = cleaned_dataset["id"].tolist().index("nonprintable")
    comment = cleaned_dataset["Comments"][nonprintable_id]
    assert comment.isprintable()


def test__proper_comment_is_retained(cleaned_dataset):
    ids = cleaned_dataset["id"].tolist()
    assert "good" in ids



"""Functions to do word-sense disambiguation using artifical contexts."""
import logging
import math

import numpy as np
from nltk import tokenize
from nltk.corpus import wordnet as wn
from scipy.spatial import distance
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score

logger = logging.getLogger(__name__)


def vectorize_aspects(aspect_counts, vectorizer):
    """Turn the aspect map into a a vector of nouns and their vector representations, which also filters aspects without embedding.

    :param aspect_counts: (dict): the dictionary with aspect counts
    :param vectorizer: (Vectorizer): the provider of word-embeddings
    :return vectors with representable aspects and their vector embeddings
    """
    aspect_vectors = []
    aspect_nouns = []
    for key, _ in aspect_counts.most_common():
        vec = vectorizer.get_vector(key)
        if vec is not None:
            aspect_nouns.append(key)
            aspect_vectors.append(vec)
    return aspect_nouns, aspect_vectors


def best_cluster(aspect_vectors):
    """
    Find the optimal cluster size using silhouette scores.

    :param aspect_vectors:  list of embeddings vectors to be clustered
    :type aspect_vectors: [:class:`numpy.array`]
    :return: the optimal number of clusters
    :rtype: int
    """
    # search a decent number of cluster-numbers for the optimum, but don't get
    # excessive
    min_cluster = 10
    max_cluster = 600
    step_cluster = 20
    safety_margin = 0.6

    silhouette_score_dict = {}
    for cluster_count in range(2, min(max_cluster, len(aspect_vectors)), step_cluster):
        kmeans_clustering = KMeans(n_clusters=cluster_count, random_state=1)
        kmeans_clustering.fit(aspect_vectors)
        silhouette_score_dict[cluster_count] = silhouette_score(
            aspect_vectors, kmeans_clustering.labels_, metric="euclidean"
        )
        logger.debug(
            "{0:d} : {1:f}".format(cluster_count, silhouette_score_dict[cluster_count])
        )
    best = sorted(
        silhouette_score_dict.items(), key=lambda item: item[1], reverse=True
    )[0][0]
    logger.debug("best cluster: {0:d}".format(best))
    logger.debug(
        sorted(silhouette_score_dict.items(), key=lambda item: item[1], reverse=True)
    )
    # return a decent default if clustering doesn't converge
    if best < min_cluster or best > safety_margin * min(
        max_cluster, len(aspect_vectors)
    ):
        return int(math.sqrt(len(aspect_vectors)))
    else:
        return best


def cluster(aspects, aspect_vectors, vectorizer):
    """Cluster aspects based on the distance of their vector representations.

    Once clusters are found, use the other aspects in a given cluster to generate the
    context for a specific aspect noun

    :param aspects: list of words for which clusters are generated
    :type aspects: [str]
    :param aspect_vectors:  list of embeddings corresponding to the the aspects
    :type aspect_vectors: [:class:`numpy.array`]
    :param vectorizer: the provider of word-embeddings for context generation
    :type vectorizer: :class:`extra_model._vectorizer.Vectorizer`
    :return: the synthetic context embedding for each of the input aspects
    :rtype: [:class:`numpy.array`]
    """
    # find the best cluster-size and run k-means clustering, resulting
    # clusters will serve as pseudo-contexts for disambiguation
    best = best_cluster(aspect_vectors)
    kmeans_clustering = KMeans(n_clusters=best, random_state=1)
    kmeans_clustering.fit(aspect_vectors)
    label_map = {}
    # map the cluster results on to the indices of the list of aspects
    for label, aspect in zip(kmeans_clustering.labels_, aspects):
        if label not in label_map:
            label_map[label] = [aspect]
        else:
            label_map[label].append(aspect)
    logger.debug(label_map)

    # generate the context for each aspect, i.e. the average of the embeddings
    # of the word in the cluster, excluding the aspect in question
    contexts = []
    for label, aspect in zip(kmeans_clustering.labels_, aspects):
        contexts.append(
            [vectorizer.get_vector(noun) for noun in label_map[label] if noun != aspect]
        )
    contexts = [np.sum(context, axis=0) for context in contexts]
    contexts = [np.divide(context, np.linalg.norm(context)) for context in contexts]

    return contexts


def match(aspect_counts, vectorizer):
    """Match a word to a specific wordnet entry, using the vector similarity of the aspects context and the synonym gloss.

    :param aspect_counts: Counter object of aspect->number of occurrence
    :type aspect_counts: :class:`collections.Counter`
    :param vectorizer:  the provider of word-embeddings for context generation
    :type vectorizer: :class:`extra_model._vectorizer.Vectorizer`
    :return list of aspects that have an embedding and best matching wordnet synonym for each aspect (can be None if no match is found)
    :rtype: ([str],[:class:`nltk.wornet.Synset`])
    """
    # prepare vector representations for clustering
    aspects, aspect_vectors = vectorize_aspects(aspect_counts, vectorizer)
    # find clusters of vectors based on the embedding similarity
    contexts = cluster(aspects, aspect_vectors, vectorizer)

    synsets = []
    for aspect in aspects:
        # find synsets for each aspect
        synset = wn.synsets(aspect.lower(), pos=wn.NOUN)
        if len(synset) == 0:
            # wordnet is missing some compounds, if that happens, we try at to
            # sea if we can have matches for constituents
            for subword in aspect.split():
                synset.extend(wn.synsets(subword.lower(), pos=wn.NOUN))
        if len(synset) == 0:
            logger.debug("No wordnet defintion found for: %s" % aspect)
        synsets.append(synset)

    # get vector embeddings for each word sense by doing a bag-of-word
    # embedding for the dictionary difeinition
    synsets_vec = [
        [synonym.definition() for synonym in synset] for synset in synsets
    ]  # get the definitions as string
    synsets_vec = [
        [tokenize.word_tokenize(synonym_def) for synonym_def in synset]
        for synset in synsets_vec
    ]  # tokenize each sting
    synsets_vec = [
        [
            [vectorizer.get_vector(word) for word in synonym_def_word]
            for synonym_def_word in synset
        ]
        for synset in synsets_vec
    ]  # gert the rmbeddding for each token
    synsets_vec = [
        [
            [embedding for embedding in synonym_def_emebd if embedding is not None]
            for synonym_def_emebd in synset
        ]
        for synset in synsets_vec
    ]  # ignore words without embedding
    synsets_vec = [
        [np.sum(synonym_embeddings, axis=0) for synonym_embeddings in synset]
        for synset in synsets_vec
    ]  # sum the embeddings to get a bag-of-words embedding
    synsets_vec = [
        [
            np.divide(synset_embedding, np.linalg.norm(synset_embedding))
            for synset_embedding in synset
        ]
        for synset in synsets_vec
    ]  # normalize

    synsets_match = []
    for noun, context, synset, synset_vec in zip(
        aspects, contexts, synsets, synsets_vec
    ):
        # filter words not represented in wordnet
        if len(synset_vec) == 0:
            synsets_match.append(None)
            continue
        # If only one word-sense is matched we don't have to disambiguate, just
        # take it
        if len(synset_vec) == 1:
            synsets_match.append(synset[0])
            continue

        # if there's ambiguity, choose the worde sense with the smallest
        # embedding distance between dictionary definition and noun context
        distances = [distance.cosine(context, synonym) for synonym in synset_vec]
        min_dist, min_index = min((val, idx) for (idx, val) in enumerate(distances))
        synsets_match.append(synset[min_index])
        logger.debug(
            "{0!s} {1!s}  at distance: {2:f}".format(noun, synset[min_index], min_dist)
        )
        logger.debug(synset[min_index].definition())

    return aspects, synsets_match



import pandas as pd
import pytest
import spacy

from extra_model._aspects import (
    acomp_list,
    adjective_list,
    adjective_negations,
    compound_noun_list,
    generate_aspects,
    parse,
)


@pytest.fixture()
def spacy_nlp():
    return spacy.load("en_core_web_sm", disable=["ner"])


def test_aspects__compound_noun_list__left_compound(spacy_nlp):
    example_text = "This is a wood screw."
    assert compound_noun_list(spacy_nlp(example_text)[4]) == ["screw", "wood screw"]


@pytest.mark.skip(reason="Left-headed compounds are exceedingly rare in English")
def test_aspects__compound_noun_list__right_compound(spacy_nlp):
    # left-headed compounds are exceedingly rare in english and in 10k
    # example texts, there was not a single one. Could remove the code or could
    # go for a deeper search for an example
    pass


def test_aspects__acomp_list(spacy_nlp):
    example_text = "The shelf is sturdy and beautiful."
    assert acomp_list(spacy_nlp(example_text)[1].head.children) == [
        "sturdy",
        "beautiful",
    ]


def test_aspects__adjective_list(spacy_nlp):
    example_text = "I bought a sturdy and beautiful shelf."
    assert adjective_list(spacy_nlp(example_text)[6].children) == [
        "sturdy",
        "beautiful",
    ]


def test_aspects__acomp_list_double_space(spacy_nlp):
    # Spacy parses double space as a token which breaks extra down the line
    example_text = "The shelf is sturdy and  beautiful."
    assert acomp_list(spacy_nlp(example_text)[1].head.children) == [
        "sturdy",
        "beautiful",
    ]


def test_aspects__adjective_list_double_space(spacy_nlp):
    # Spacy parses double space as a token which breaks extra down the line
    example_text = "I bought a sturdy and  beautiful shelf."
    assert adjective_list(spacy_nlp(example_text)[7].children) == [
        "sturdy",
        "beautiful",
    ]


@pytest.mark.skip(reason="This needs to be looked at when fixing negations")
def test_aspects__adjective_negations__direct(spacy_nlp):
    example_text = "This not so sturdy table is a disappointment."
    assert adjective_negations(spacy_nlp(example_text)[1]) == ["sturdy"]
    # There is a difference here in spacy versions that will need to be investigated.
    # succeeds in 2.0.18 but fails in 3.0.


def test_aspects__adjective_negations__right_non_attr(spacy_nlp):
    example_text = "This color is not pretty."
    assert adjective_negations(spacy_nlp(example_text)[3]) == ["pretty"]


def test_aspects__adjective_negations__right_attr(spacy_nlp):
    example_text = "This is not a terrible table."
    assert adjective_negations(spacy_nlp(example_text)[2]) == ["terrible"]


def test_aspects__parse_multiple_spaces(spacy_nlp, mocker):
    # latest version of spacy parse whitespace as a separate token which sometimes picked up as an adjective
    # we want to make sure that no empty adjectives make it further after parsing
    example_texts = [
        "Wonderful quality - ease in ordering and returning                                     great prices",
        "I bought a sturdy and  beautiful shelf.",
        "Wonderful quality - ease in ordering and returning                             great prices",
    ]
    data_frame = pd.DataFrame(example_texts, columns=["Comments"])
    mocker.patch("spacy.load", return_value=spacy_nlp)
    result = parse(data_frame)
    descriptors = result["descriptor"].tolist()
    assert all([desc.strip() != "" for desc in descriptors])


def test_aspects__parse(spacy_nlp, mocker):
    # chose a text that exercise as much code as possible
    # second part of the sentence is here to check that negations are properly filtered
    example_text = "The wooden cabinet serves its purpose, it's not terrible."
    data_frame = pd.DataFrame([{"Comments": example_text}])
    mocker.patch("spacy.load", return_value=spacy_nlp)
    result = parse(data_frame)
    assert (
        result.iloc[0]["CiD"] == 0
        and result.iloc[0]["position"] == 11
        and result.iloc[0]["aspect"] == "cabinet"
        and result.iloc[0]["descriptor"] == "wooden"
        and not result.iloc[0]["is_negated"]
    )


def test_aspects_generate_aspects(spacy_nlp, mocker):
    example_text = "The wooden cabinet serves its purpose, it's not terrible."
    data_frame = pd.DataFrame([{"Comments": example_text}])
    mocker.patch("spacy.load", return_value=spacy_nlp)
    result = generate_aspects(data_frame)
    assert (
        result.iloc[0]["CiD"] == 0
        and result.iloc[0]["position"] == 11
        and result.iloc[0]["aspect"] == "cabinet"
        and result.iloc[0]["descriptor"] == "wooden"
        and not result.iloc[0]["is_negated"]
    )



import os

import pytest
from gensim.models import KeyedVectors

from extra_model._vectorizer import Vectorizer


@pytest.fixture()
def vec():
    # preprocess plain-text test embeddings to proper binary format for vectorizer
    glove_file = "tests/resources/test_adjectives.vec"
    prepro_file = "tests/resources/test_adjectives.prepro"
    model = KeyedVectors.load_word2vec_format(glove_file, binary=False, no_header=True)
    model.save(prepro_file)

    yield Vectorizer(prepro_file)

    # cleanup
    os.remove(prepro_file)


def test__word_exists(vec):
    res = vec.get_vector("little")
    assert res is not None


def test__word_doesnt_exist(vec):
    res = vec.get_vector("gibberrish")
    assert res is None


def test__subword_doesnt_exist(vec):
    res = vec.get_vector("gibberish little")
    assert res is None


def test__compound_word_exists(vec):
    res = vec.get_vector("little small")
    assert res is not None



import datetime as dt
import json
import logging
import os
import shutil
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd

from extra_model._adjectives import adjective_info
from extra_model._aspects import generate_aspects
from extra_model._filter import filter
from extra_model._summarize import link_aspects_to_texts, link_aspects_to_topics, qa
from extra_model._topics import get_topics
from extra_model._vectorizer import Vectorizer

CB_BASE_DIR = "/"
EMBEDDING_TYPE = "glove.840B.300d"


logger = logging.getLogger(__name__)


class ModelBase:
    """Base class that provides file loading functionality."""

    models_folder: str
    _storage_metadata: Dict[str, str]

    def load_from_files(self):
        """Load model files."""
        # load the storage metadata info obtained when loading embeddings from model storage
        file_name = os.path.join(self.models_folder, "metadata.json")
        storage_metadata = {}
        if os.path.isfile(file_name):
            with open(file_name) as f:
                storage_metadata = json.load(f)

            # overwrite storage_metadata for select fields
            self._storage_metadata["id"] = storage_metadata.get("id")
            self._storage_metadata["dag_id"] = storage_metadata.get("dag_id")
            self._storage_metadata["dag_run_id"] = storage_metadata.get("dag_run_id")
            self._storage_metadata["date_trained"] = storage_metadata.get(
                "date_trained"
            )
            self._storage_metadata["target_training_date"] = storage_metadata.get(
                "target_training_date"
            )
            self._storage_metadata["json_extras"] = storage_metadata.get("json_extras")


class ExtraModelBase:
    """Extra model class that provides an interface for training and predicting."""

    is_trained = False
    models_folder = "/embeddings"
    training_folder = ""

    _filenames = {
        "embeddings": f"{EMBEDDING_TYPE}.vectors.npy",
        "prepro": EMBEDDING_TYPE,
    }
    # there is no need for this since Extra doesn't create any artifacts
    _training_artifacts: Dict[str, str] = {}

    def __init__(
        self,
        dag_id="",
        dag_run_id="",
        models_folder=models_folder,
        embedding_type=EMBEDDING_TYPE,
    ):
        """Init function for ExtraModel object.

        :param dag_id: Name of dag
        :param dag_runs_ids: Dag run IDs
        :param models_folder: Path to folder where model files are stored
        :param embedding_type: Name of embedding file. Default is "glove.840B.300d"
        """
        self.models_folder = models_folder
        self.embedding_type = embedding_type
        self.api_spec_names = {
            "position": "Position",
            "aspect": "Aspect",
            "descriptor": "Descriptor",
            "aspect_count": "AspectCount",
            "wordnet_node": "WordnetNode",
            "sentiment_compound_aspect": "SentimentCompound",
            "sentiment_binary_aspect": "SentimentBinary",
            "adcluster": "AdCluster",
            "source_guid": "CommentId",
            "topic": "Topic",
            "importance": "TopicImportance",
            "sentiment_compound_topic": "TopicSentimentCompound",
            "sentiment_binary_topic": "TopicSentimentBinary",
            "num_occurance": "TopicCount",
        }

        self._storage_metadata = {
            "type": "text",
            "owner": "blank",
            "description": "Running ExtRA algorithm",
            "display_name": "extra-model",
            "features": {},
            "hyperparameters": {},
            "dag_id": dag_id,
            "dag_run_id": dag_run_id,
            "is_scheduled_creation": False,
            "date_trained": str(dt.date.today()),
            "target_training_date": str(dt.date.today()),
            "json_extras": {"classification_report_json": ""},
        }
        for key in self._filenames:
            self._storage_metadata[key] = {}

    def storage_metadata(self):
        """Docstring."""
        return self._storage_metadata

    def load_from_files(self):
        """Docstring."""
        super().load_from_files()
        self.vectorizer = Vectorizer(
            os.path.join(self.models_folder, self.embedding_type)
        )
        self.is_trained = True

    def train(self):
        """Docstring."""
        for key, filename in self._filenames.items():
            logger.debug(f"Downloading {key}")
            shutil.copyfile(
                src=os.path.join(CB_BASE_DIR, filename),
                dst=os.path.join(self.models_folder, filename),
            )
        self.is_trained = True

    def predict(self, comments: List[Dict[str, str]]) -> List[Dict]:
        """Docstring."""
        if not self.is_trained:
            raise RuntimeError("Extra must be trained before you can predict!")
        dataframe_texts = pd.DataFrame(comments)
        dataframe_texts.rename(
            {"CommentId": "source_guid"}, axis="columns", inplace=True
        )
        dataframe_texts = filter(dataframe_texts)
        dataframe_aspects = generate_aspects(dataframe_texts)

        if dataframe_aspects.empty:
            raise ValueError(
                "Input dataset doesn't contain valid aspects, stopping the algorithm"
            )

        # aggregate and abstract aspects into topics
        dataframe_topics = get_topics(dataframe_aspects, self.vectorizer)
        dataframe_topics, dataframe_aspects = adjective_info(
            dataframe_topics, dataframe_aspects, self.vectorizer
        )
        dataframe_aspects = link_aspects_to_topics(dataframe_aspects, dataframe_topics)
        dataframe_aspects = link_aspects_to_texts(dataframe_aspects, dataframe_texts)

        # do some extra book-keeping if debug-level is set low enough
        if logger.isEnabledFor(20):
            qa(dataframe_texts, dataframe_aspects, dataframe_topics)

        # write output_tables, after dropping auxilliary information
        dataframe_topics.loc[:, "num_occurance"] = dataframe_topics["rawnums"].apply(
            lambda counts: sum(counts)
        )
        dataframe_topics = dataframe_topics[
            [
                "topicID",
                "topic",
                "importance",
                "sentiment_compound",
                "sentiment_binary",
                "num_occurance",
            ]
        ]

        dataframe_aspects.dropna(axis=0, inplace=True)
        dataframe_aspects["topicID"] = dataframe_aspects["topicID"].astype(int)

        output = dataframe_aspects.merge(
            dataframe_topics, on="topicID", suffixes=("_aspect", "_topic")
        )
        return standardize_output(output, names=self.api_spec_names).to_dict("records")


# NOTE: improve typehints!
def extra_factory(bases: Optional[Union[Any, Tuple[Any]]] = None) -> Any:
    """Create for ExtraModel class types.

    Will dynamically create the class when called with the provided base classes.

    :param bases: Base classes to be used when creating ExtraModel class
    :type bases: Class type or tuple of class types
    :return: ExtraModel class
    """
    if bases is None:
        bases = (ModelBase,)
    elif not isinstance(bases, tuple):
        bases = (bases,)
    bases = (ExtraModelBase,) + bases

    return type("ExtraModel", bases, {})


def standardize_output(data: pd.DataFrame, names: dict) -> pd.DataFrame:
    """Standarize output.

    Ensures the following:
    - only required columns are returned and
    - they are named according to spec

    :param data: input dataframe.
    :param names: dictionary to standardize output to API spec.
    :return: renamed dataframe.
    """
    return data[list(names.keys())].rename(columns=names)


ExtraModel = extra_factory()



"""Fill out and link the dataframes for topics, aspects and texts. Provide some summary ouput if debug-level is set to info or lower."""
import logging

logger = logging.getLogger(__name__)


def qa(dataframe_texts, dataframe_aspects, dataframe_topics):
    """Print summary information.

    :param dataframe_texts: dataframe with the raw texts (for example output)
    :type dataframe_texts: :class:`pandas.DataFrame`
    :param dataframe_aspects:  dataframe with the aspects
    :type dataframe_texts: :class:`pandas.DataFrame`
    :param dataframe_topics: dataframe with the topics
    :type dataframe_texts: :class:`pandas.DataFrame`
    """
    for ind in dataframe_topics.index:
        logger.debug("\n++++++++++++++++++++++++++++++\n")
        logger.debug(
            "analyzing topic {0!s} of importance {1:f}".format(
                dataframe_topics["topic"][ind], dataframe_topics["importance"][ind]
            )
        )
        logger.debug(
            "with seniment {0:f}(compound) / {1:f}(binary)\n".format(
                dataframe_topics["sentiment_compound"][ind],
                dataframe_topics["sentiment_binary"][ind],
            )
        )
        # attached terms by raw number
        logger.debug("Common terms (by raw number)")
        for term, num in zip(
            dataframe_topics["rawterms"][ind], dataframe_topics["rawnums"][ind]
        ):
            logger.debug("{0!s} ({1:d}) ".format(term, num))

        # count attached terms by distance weighted number XXX this section will
        # go after debugginf
        logger.debug("Common terms for aspect (distance weighted)")
        for term, num in zip(
            dataframe_topics["weightedterms"][ind], dataframe_topics["weights"][ind]
        ):
            logger.debug("{0!s} ({1:f}) ".format(term, num))

        # adjective clusters
        if dataframe_topics["adjective_clusters"][ind] is not None:
            logger.debug("\nFound adjective clusters:\n")
            representatives, sizes, adjectives = dataframe_topics["adjective_clusters"][
                ind
            ]
            for representative, size, subads in zip(representatives, sizes, adjectives):
                logger.debug(
                    "For representative: {}({}): ".format(representative, size)
                )
                adstring = ""
                for ad, count in subads:
                    adstring += "{}({}), ".format(ad, count)
                logger.debug(adstring)

            # representative comments (most common term  and most commen
            # adjective)
            logger.debug(
                "\nrepresentative comments (most common term with most common adjective cluster)"
            )
            comments = dataframe_aspects[
                dataframe_aspects["aspect"].isin([dataframe_topics["rawterms"][ind][0]])
                & dataframe_aspects["adcluster"].isin(
                    [dataframe_topics["adjective_clusters"][ind][0][0]]
                )
            ]
            comments = comments.head(5)
            for oi in comments["CiD"]:
                logger.debug(dataframe_texts["Comments"][oi])


def set_aspect(topic, dataframe_aspects):
    """For a given topic, set topic and adjective cluster fields in the aspect_dataframe.

    :param topic: the topic and it's associated information that we need to copy to the relevant entries in the aspect frame
    :type topic: :class:`pandas.DataFrame.Row`
    :param dataframe_aspects: the dataframe to be enriched with topic information
    :type dataframe_aspects: :class:`pandas.DataFrame`
    """
    dataframe_aspects.loc[
        dataframe_aspects["aspect"].isin(topic["rawterms"]), "topicID"
    ] = topic["topicID"]

    # this can happen if there is only a single adjective
    if topic["adjective_clusters"] is not None:
        for representative, all_adjectives in zip(
            topic["adjective_clusters"][0], topic["adjective_clusters"][2]
        ):
            dataframe_aspects.loc[
                dataframe_aspects["aspect"].isin(topic["rawterms"])
                & dataframe_aspects["descriptor"].isin(list(zip(*all_adjectives))[0]),
                "adcluster",
            ] = representative


def link_aspects_to_topics(dataframe_aspects, dataframe_topics):
    """Fill topic and adjective cluster information into the aspect dataframe.

    :param dataframe_aspects: the dataframe to be enriched
    :type dataframe_aspects: :class:`pandas.DataFrame`
    :param dataframe_topics:  the dataframe that has the topic and adjective cluster information
    :type dataframe_topics: :class:`pandas.DataFrame`
    :return: the enriched dataframe
    :rtype: :class:`pandas.DataFrame`
    """
    # create empty collumns to be filled
    dataframe_aspects["topicID"] = None
    dataframe_aspects["adcluster"] = None

    # create a topic-id from the dataframe index
    dataframe_topics.reset_index(inplace=True)
    dataframe_topics.rename({"index": "topicID"}, axis="columns", inplace=True)

    dataframe_topics.apply(lambda topic: set_aspect(topic, dataframe_aspects), axis=1)

    return dataframe_aspects


def link_aspects_to_texts(dataframe_aspects, dataframe_texts):
    """Transfer the original text identifier from the original text data table into the final aspect table.

    :param dataframe_aspects: table to be enriched
    :type dataframe_aspects: :class:`pandas.DataFrame`
    :param dataframe_texts: original table from which this information is extracted
    :type dataframe_texts: :class:`pandas.DataFrame`
    :return: The aspect dataframe which has been enriched with source-information
    :rtype: :class:`pandas.DataFrame`
    """
    # we don't need everything from the original, let's keep only the useful
    # info to safe space on the output table
    keepers = dataframe_aspects.columns.values.tolist()
    keepers.append("source_guid")  # unique identifier for the utterance
    logger.debug("keeping aspects columns: {}".format(keepers))

    # join aspects to the original texts
    dataframe_aspects = dataframe_aspects.join(dataframe_texts, on="CiD")

    return dataframe_aspects[keepers]



import logging

import numpy as np
from gensim.models import KeyedVectors

logger = logging.getLogger(__name__)


class Vectorizer:
    """Simple Vectorizer class using pre-trained vectors."""

    def __init__(self, embedding_file):
        """
        Use the generic gensim vector embedding lookup.

        Currently using pretrained glove embeddings, but anything goes.
        :param embedding_file: pathname for the file that stores the word-embeddings in gensim keyed-vectors format
        :type str
        """
        self.wv_glove = KeyedVectors.load(embedding_file, mmap="r")

    def get_vector(self, key):
        """
        Return the vector embedding for a given word.

        According to the following logic:
            - if no embedding is found for this word, check if it's a compound
            - if it's a compound try to take the average embedding of the constituent words
            - if a that fails too, return None, downstream code will have to deal with the absence of an embedding
        (typical course of action is to ignore such a term)
        :param key: the word to be embedded
        :type str
        :return: the embedding vector. Number of dimensions is set by the input file
        :rtype np.array:
        """
        if key.lower() in self.wv_glove.key_to_index:
            return self.wv_glove.get_vector(key.lower(), norm=True)
        else:
            for subword in key.split():
                if subword.lower() not in self.wv_glove.key_to_index:
                    logger.debug("can't vectorize {0!s}".format(key))
                    return None
            res = np.sum(
                [
                    self.wv_glove.get_vector(subword.lower(), norm=True)
                    for subword in key.split()
                ],
                axis=0,
            )
            norm = np.linalg.norm(res)
            return np.divide(res, norm)



import math
import os
from collections import Counter

import pytest
from gensim.models import KeyedVectors

from extra_model._disambiguate import best_cluster, cluster, match, vectorize_aspects
from extra_model._vectorizer import Vectorizer


@pytest.fixture()
def vec_cluster():
    # preprocess plain-text test embeddings to proper binary format for vectorizer
    glove_file = "tests/resources/test_disambiguate.vec"
    prepro_file = "tests/resources/test_disambiguate.prepro"
    model = KeyedVectors.load_word2vec_format(glove_file, binary=False, no_header=True)
    model.save(prepro_file)

    yield Vectorizer(prepro_file)

    # cleanup
    os.remove(prepro_file)


@pytest.fixture()
def vec():
    return Vectorizer("tests/resources/small_embeddings")


def test__vectorize_aspects(vec_cluster):
    counts = Counter(["chair", "chair", "table", "unkown"])
    nouns, vecs = vectorize_aspects(counts, vec_cluster)
    assert (
        nouns[0] == "chair"  # should be ordered by fequency, so chair first
        and nouns[1] == "table"
        and (vecs[0] == vec_cluster.get_vector("chair")).all()
        and (vecs[1] == vec_cluster.get_vector("table")).all()
        and len(vecs) == 2  # aggregate identical words and drop unknown
        and len(nouns) == 2
    )


def test__best_cluster__too_few_crash():
    # there need to be at least three vectors for the clustering to make sense
    with pytest.raises(IndexError):
        best_cluster([[1.0, 0.0], [0.0, 0.5]])


def test__best_cluster__too_few_default():
    # for a very low number of vectors (<10) should default to int(sqrt(len(vectors)))
    assert best_cluster([[1.0, 0.0], [0.0, 0.5], [-0.5, -0.5]]) == 1


def test__best_cluster():
    # a ~realistic clusterting example
    vectors = []
    for i in range(100):
        vectors.append(
            [math.sin(math.pi * 2 * i / 100), math.cos(math.pi * 2 * i / 100)]
        )
    assert best_cluster(vectors) == 22


def test__cluster(vec_cluster, mocker):
    # we don't care about cluster size optimization for the test, just make sure we get the conexts aggregated right.
    mocker.patch("extra_model._disambiguate.best_cluster", return_value=2)
    aspects = ["table", "chair", "arm", "leg"]
    aspect_vectors = [vec_cluster.get_vector(aspect) for aspect in aspects]

    contexts = cluster(aspects, aspect_vectors, vec_cluster)
    # example is chosen such that table<->chair and arm<->leg should be each others context
    assert (
        (contexts[0] == aspect_vectors[1]).all()
        and (contexts[1] == aspect_vectors[0]).all()
        and (contexts[2] == aspect_vectors[3]).all()
        and (contexts[3] == aspect_vectors[2]).all()
    )


def test__match(vec):
    # mostly an integration test.
    aspects, synsets = match(
        # use distinct numbers of terms here to obtain definite output ordering (most common first)
        Counter(
            [
                "chair",
                "chair",
                "chair",
                "chair",
                "sofa",
                "sofa",
                "sofa",
                "table",
                "table",
                "asdf",
            ]
        ),
        vec,
    )
    assert (
        aspects[0] == "chair"
        and aspects[1] == "sofa"
        and aspects[2] == "table"
        and aspects[3] == "asdf"
        and synsets[0].name() == "chair.n.01"
        and synsets[1].name() == "sofa.n.01"
        and synsets[2].name() == "table.n.03"  # disambiguation failure
        and not synsets[3]  # no synset for the unknown word
    )



import logging

import pandas as pd
import spacy
from spacy.symbols import NOUN, VERB, acomp, amod, nsubj

"""Generate the basic phrases that will be used for clustering
Major steps:
1) Run Spacy for POS tags
2) extract NPs described by adjectives
3) filter for "high quality phrases" using the Autophrase tool
"""
logger = logging.getLogger(__name__)


def compound_noun_list(token):
    """Find compound nouns.

    :param token: token for which to generate potential compound nouns
    :type token: :class:`spacy.token`
    :return: list of potential compounds
    :rtype: [string]
    """
    nouns = [token.text]
    for nc in token.lefts:
        if nc.dep_ == "compound":
            nouns.append(nc.text + " " + token.text)
    for nc in token.rights:
        if nc.dep_ == "compound":
            nouns.append(token.text + " " + nc.text)
    return nouns


def acomp_list(tokens):
    """Find descriptions for a given token.

    :param tokens: list of tokens that are children of the head of the noun for which descriptions are searched.
    :type tokens: [:class:`spacy.token`]
    :return: list of adjectives
    :rtype: [string]
    """
    acomps = []
    for child in tokens:
        if child.dep == acomp:
            acomps.append(child.text)
            for grandchild in child.children:
                # find both X and Y in patterns of the form "product is X and Y"
                if grandchild.dep_ == "conj" and not grandchild.is_space:
                    # grandchild.is_space is handling whitespace cases
                    acomps.append(grandchild.text)
    return acomps


def adjective_list(tokens):
    """Find adjectives modifying a given noun.

    :param tokens: tokens of potential adjectice candidates (children of the noun and children of the head for compounds)
    :type tokens: [:class:`spacy.token`]
    :return: list of adjectives
    :rtype: [string]
    """
    adjectives = []
    for child in tokens:
        if child.dep == amod:
            adjectives.append(child.text)
            for grandchild in child.children:
                # find both X and Y in patterns of the form "the X and Y product"
                if grandchild.dep_ == "conj" and not grandchild.is_space:
                    # grandchild.is_space is handling whitespace cases
                    adjectives.append(grandchild.text)
    return adjectives


def adjective_negations(token):
    """Find all negated adjectives in a sentence.

    :param token: negation token to handle
    :type token: :class:`spacy.token`
    :return: list of negated adjectives
    :rtype: [string]
    """
    negated_adjectives = []

    # "This color is not pretty" case
    if token.head.dep_ == "conj" or token.head.dep_ == "amod":
        negated_adjectives.append(token.head.text)
    # always looking to the right -- not handling sarcasm (e.g., "Beautiful
    # table, not!")
    for tok in token.head.rights:
        if tok.dep_ == "conj" or tok.dep_ == "amod" or tok.dep_ == "acomp":
            negated_adjectives.append(tok.text)
        # noun phrase case "This is not a terrible table"
        if tok.dep_ == "attr":
            for n_tok in tok.children:
                if (
                    n_tok.dep_ == "conj"
                    or n_tok.dep_ == "amod"
                    or n_tok.dep_ == "acomp"
                ):
                    negated_adjectives.append(n_tok.text)

    return negated_adjectives


def parse(dataframe_texts):  # noqa: C901
    """Parse the comments and extract a list of potential aspects based on grammatical relations.

    (e.g. modified by adjective)

    :param dataframe_texts: a dataframe with the raw texts. The collumn wit the texts needs to be called 'Comments'
    :type dataframe_texts: :class:`pandas.DataFrame`
    :return: a dataframe with the aspect candidates
    :rtype: :class:`pandas.DataFrame`
    """
    nlp = spacy.load("en_core_web_sm", disable=["ner"])
    # make a new dataframe with one row for each aspect/adjective pair
    rowlist = []

    # loop over all the texts and do syntax analysis, keeping valid nouns+adjectived
    # runs faster by running spacy in batch-mode
    # we need to keep the index here in order to be able to connect back to
    # the original text later

    # n_threads > 5 can segfault with long (>500 tokens) sentences
    # n_threads has been deprecated in spacy 3.x - https://spacy.io/usage/v2-1#incompat
    for index, document in zip(
        dataframe_texts.index,
        nlp.pipe(dataframe_texts.Comments, batch_size=500),
    ):
        negated_adjectives = []
        for token in document:
            nouns = compound_noun_list(token)
            adjectives = []
            if token.dep == nsubj and token.pos == NOUN and token.head.pos == VERB:
                # find nouns with descriptions
                adjectives.extend(acomp_list(token.head.children))
            if token.pos == NOUN:  # find nouns with adjectives
                adjectives.extend(adjective_list(token.children))
                # necessary for compound nouns
                adjectives.extend(adjective_list(token.head.children))
            adjectives = list(dict.fromkeys(adjectives))  # remove duplicates
            adjectives = list(filter(lambda adj: adj.strip() != "", adjectives))
            if len(adjectives) != 0:
                # since negation can come much later in the sentence, finding
                # it here
                for tok in document:
                    if tok.dep_ == "neg":
                        negated_adjectives = adjective_negations(tok)
                for noun in nouns:
                    for adjective in adjectives:
                        row = {
                            "CiD": index,  # index into the initial raw text dataframe
                            # position within the text of the candidate noun
                            # (in letters!)
                            "position": token.idx,
                            "aspect": noun,
                            "descriptor": adjective,
                            "is_negated": adjective in negated_adjectives,
                        }

                        rowlist.append(row)

    dataframe_aspects = pd.DataFrame(
        rowlist, columns=["CiD", "position", "aspect", "descriptor", "is_negated"]
    )
    return dataframe_aspects


def generate_aspects(dataframe_texts):
    """Generate the aspects that will be merged into topics from the raw texts.

    :param dataframe_texts: a dataframe with the raw texts in the column 'Comments'
    :type dataframe_texts: :class:`pandas.DataFrame`
    :return: a dataframe with the aspect candidates, their associated description, index of original text in the
    input dataframe and location of word in the text
    :rtype: :class:`pandas.DataFrame`
    """
    logger.debug(dataframe_texts.head())
    # extract candidate noun-phrases using spacy
    dataframe_aspects = parse(dataframe_texts)
    return dataframe_aspects



import subprocess
from pathlib import Path

import pytest

from extra_model._errors import ExtraModelError
from extra_model._setup import URL, cleanup, download_file, format_file, run_subprocess
from extra_model._setup import setup as setup_extra
from extra_model._setup import unzip_file

OUTPUT = "/output"


@pytest.fixture
def subprocess_run_mock(mocker):
    # NOTE: run function in subprocess module
    return mocker.patch("subprocess.run")


@pytest.fixture
def run_subprocess_mock(mocker):
    # NOTE: our wrapper around subprocess.run
    return mocker.patch("extra_model._setup.run_subprocess")


@pytest.fixture
def input_mock(mocker):
    return mocker.patch("builtins.input", return_value="y")


@pytest.fixture
def download_file_mock(mocker):
    return mocker.patch("extra_model._setup.download_file")


@pytest.fixture
def unzip_file_mock(mocker):
    return mocker.patch("extra_model._setup.unzip_file")


@pytest.fixture
def format_file_mock(mocker):
    return mocker.patch("extra_model._setup.format_file")


@pytest.fixture
def cleanup_mock(mocker):
    return mocker.patch("extra_model._setup.cleanup")


@pytest.fixture
def create_output_path(mocker):
    """Mock for Path truediv (/) operator"""

    def _create_output_path(file_exists=True):
        output_file = mocker.Mock()
        output_file.is_file.return_value = file_exists
        output_path = mocker.MagicMock()
        output_path.__truediv__.return_value = output_file
        return output_path

    return _create_output_path


def test_setup__input_return_n__setup_functions_not_called(
    input_mock, download_file_mock, unzip_file_mock, format_file_mock, cleanup_mock
):

    input_mock.return_value = "n"

    setup_extra(OUTPUT)

    download_file_mock.assert_not_called()
    unzip_file_mock.assert_not_called()
    format_file_mock.assert_not_called()
    cleanup_mock.assert_not_called()


def test_setup__input_return_True__setup_functions_called(
    input_mock, download_file_mock, unzip_file_mock, format_file_mock, cleanup_mock
):

    input_mock.return_value = "y"

    setup_extra(OUTPUT)

    download_file_mock.assert_called_once()
    unzip_file_mock.assert_called_once()
    format_file_mock.assert_called_once()
    cleanup_mock.assert_called_once()


def test_setup__output_path_set__argument_passed_to_setup_functions(
    input_mock, download_file_mock, unzip_file_mock, format_file_mock, cleanup_mock
):

    setup_extra(OUTPUT)

    download_file_mock.assert_called_once_with(URL, OUTPUT)

    unzip_file_mock.assert_called_once_with(download_file_mock.return_value, OUTPUT)

    format_file_mock.assert_called_once_with(unzip_file_mock.return_value, OUTPUT)

    cleanup_mock.assert_called_once_with(
        [download_file_mock.return_value, unzip_file_mock.return_value]
    )


def test_download_file__output_file_found__skip_download_file(
    create_output_path, run_subprocess_mock
):

    output_path = create_output_path(file_exists=True)

    download_file(URL, output_path)

    run_subprocess_mock.assert_not_called()


def test_download_file__output_file_missing__download_file(
    create_output_path, run_subprocess_mock
):

    output_path = create_output_path(file_exists=False)

    download_file(URL, output_path)

    run_subprocess_mock.assert_called_once()


def test_unzip_file__output_file_found__skip_unzip_file(
    create_output_path, run_subprocess_mock
):

    file = Path("path/to/file.zip")
    output_path = create_output_path(file_exists=True)

    unzip_file(file, output_path)

    run_subprocess_mock.assert_not_called()


def test_unzip_file__output_file_missing__unzip_file(
    create_output_path, run_subprocess_mock
):

    file = Path("path/to/file.zip")
    output_path = create_output_path(file_exists=False)

    unzip_file(file, output_path)

    run_subprocess_mock.assert_called_once()


def test_format_file__output_file_found__skip_format_file(mocker, create_output_path):

    datapath_mock = mocker.patch("extra_model._setup.datapath")
    mocker.patch("extra_model._setup.KeyedVectors")

    file = Path("path/to/file.txt")
    output_path = create_output_path(file_exists=True)

    format_file(file, output_path)

    datapath_mock.assert_not_called()


def test_format_file__output_file_missing__format_file(mocker, create_output_path):

    datapath_mock = mocker.patch("extra_model._setup.datapath")
    mocker.patch("extra_model._setup.KeyedVectors")

    file = Path("path/to/file.txt")
    output_path = create_output_path(file_exists=False)

    format_file(file, output_path)

    datapath_mock.assert_called_once()


def test_cleanup__files_unliked(mocker):
    file = mocker.Mock()
    cleanup([file])
    file.unlink.assert_called_once()


def test_run_subprocess__subprocess_called_with_command(subprocess_run_mock):
    command = "command"
    run_subprocess([command])
    subprocess_run_mock.assert_called_once_with(command, check=True, shell=True)


def test_run_subprocess__return_code_not_zero__raise_ExtraModelError(
    subprocess_run_mock,
):
    subprocess_run_mock.side_effect = subprocess.CalledProcessError(1, "command")

    with pytest.raises(ExtraModelError, match=r"Subprocess error: ."):
        run_subprocess("command")



import logging
import sys
from pathlib import Path

import click

from extra_model._errors import ExtraModelError
from extra_model._run import run
from extra_model._setup import setup

logger = logging.getLogger(__name__)


OUTPUT_PATH = "/io"
OUTPUT_FILENAME = "result.csv"
EMBEDDINGS_PATH = "/embeddings"


@click.command()
@click.argument("input_path", type=Path)
@click.argument("output_path", type=Path, default=OUTPUT_PATH)
@click.argument("output_filename", type=Path, default=OUTPUT_FILENAME)
@click.argument("embeddings_path", type=Path, default=EMBEDDINGS_PATH)
@click.option("--debug", is_flag=True, help="Enable debug logging")
def entrypoint(
    input_path: Path,
    output_path: Path,
    output_filename: Path,
    embeddings_path: Path,
    debug: bool = False,
) -> None:
    """Run the Extra algorithm for unsupervised topic extraction.

    INPUT_PATH is the path to the input parquet file with the user generated texts.

    OUTPUT_PATH is the path to the output directory. Default is `/io`.

    OUTPUT_FILENAME is the filename of the output file. Default is `result.csv`.
    The `.csv` file extension is not enforced. Please take care of this accordingly.

    EMBEDDINGS_PATH is the path where the extra model will load the embeddings from.
    defaults to `./embeddings`.
    """
    logging.getLogger("extra_model").setLevel("DEBUG" if debug else "INFO")

    try:
        run(input_path, output_path, output_filename, embeddings_path)
        sys.exit(0)

    except ExtraModelError as e:
        logger.exception(e) if debug else logger.error(e)
        sys.exit(1)


@click.command()
@click.argument("output_path", type=Path, default=EMBEDDINGS_PATH)
def entrypoint_setup(output_path: Path) -> None:
    """Download resources.

    Will download and format glove embeddings.

    OUTPUT_PATH is the path to the output directory. Default is `/embeddings`.
    """
    try:
        setup(output_path)
        sys.exit(0)

    except ExtraModelError as e:
        logger.error(e)
        sys.exit(1)



"""Cluster adjectives and extract sentiment."""

from collections import Counter

import numpy as np
from nltk.corpus import wordnet as wn
from sklearn.neighbors import BallTree
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer


def cluster_adjectives(adjective_counts, vectorizer):  # noqa: C901
    """Cluster adjectives based on a constant radius clustering algorithm.

    Technical implementation uses a scikitlearn BallTree.

    :param adjective_counts: dictionary with adjectives and their counts
    :type adjective_counts: [(str,int)]
    :param vectorizer:  provide embeddings to evaluate adjective similarity
    :type vectorizer: :class:`_vectorizer.vectorizer`
    :return: first list contains the representative words for the adjective clusters,
        second list the size (total number of clustered adjectives) for each cluster,
        third list has for each cluster a list of the constituent adjectives with number of occurence
    :rtype: ([string],[int],[[(string,int)]])
    """
    adjectives, counts = zip(*adjective_counts)
    adjectives = list(adjectives)
    counts = list(counts)

    # find vector embeddings for the adjectives and filter adjectives for which ew don't have an embedding
    # most of these are typos
    vectors = [vectorizer.get_vector(ad) for ad in adjectives]
    for i in range(len(adjectives)):
        if vectors[i] is None:
            adjectives[i] = None
            counts[i] = None
    adjectives = [ad for ad in adjectives if ad is not None]
    counts = [count for count in counts if count is not None]
    vectors = [vec for vec in vectors if vec is not None]
    # keep track of indices for further processing
    indices = [i for i in range(len(adjectives))]
    clusters = []
    cluster_size = []
    # use the adjective closest to the cluster center as representative
    cluster_representative = []
    cluster_ads = []

    if (
        len(vectors) == 0
    ):  # no embeddable adjective for this topic -> there is no cluster to build
        return
    if len(vectors) == 1:  # Balltree doesn't work if list contains a single vector
        # in this case, the single word is the whole cluster
        return (adjectives, [counts], [[(adjectives[0], counts[0])]])
    tree = BallTree(vectors)
    working_set = list(zip(adjectives, indices))
    already_clustered = set()
    # run the constant radius clustering:
    # as long as there are unclustered adjectives left, take the one with the highes count
    # and cluster all other words within a given distance into it.
    while len(working_set) > 0:
        # radius of 0.8 hand-tuned to give decent adjective clusters
        cluster_indices = tree.query_radius([vectors[working_set[0][1]]], 0.8)
        new_cluster = []
        for i in cluster_indices[0]:
            new_cluster.append(i)
        new_cluster = [index for index in new_cluster if index not in already_clustered]
        # clean the new cluster from opposites:
        # get the synset for the most common word in the cluster
        master_synset = wn.synsets(working_set[0][0].lower(), pos=wn.ADJ)
        master_antonyms = set()
        for syn in master_synset:
            for lemma in syn.lemmas():
                master_antonyms.update(lemma.antonyms())
        for i in range(len(new_cluster)):
            # skip the most common word, in case it is among its own antonyms
            # (shouldn't happen)
            if new_cluster[i] == working_set[0][1]:
                continue
            for syn in wn.synsets(adjectives[new_cluster[i]].lower(), pos=wn.ADJ):
                for lemma in syn.lemmas():
                    if lemma in master_antonyms:
                        new_cluster[i] = None
                        break
                if new_cluster[i] is None:
                    break
        new_cluster = [ind for ind in new_cluster if ind is not None]
        already_clustered.update(new_cluster)
        for i in range(len(working_set)):
            if working_set[i][1] in already_clustered:
                working_set[i] = None
        working_set = [it for it in working_set if it is not None]
        clusters.append(new_cluster)
        cluster_size.append(sum(counts[i] for i in new_cluster))
        cluster_ads.append([(adjectives[i], counts[i]) for i in new_cluster])

        # get the representative as closest word to the cluster-mean
        cluster_weights = [counts[i] for i in new_cluster]
        cluster_vecs = [vectors[i] for i in new_cluster]
        cluster_representative.append(
            adjectives[
                tree.query(
                    [np.average(cluster_vecs, axis=(0), weights=cluster_weights)],
                    return_distance=False,
                )[0][0]
            ]
        )

    # sort clusters by size before returning
    cluster_size, cluster_ads, cluster_representative = zip(
        *sorted(
            zip(cluster_size, cluster_ads, cluster_representative),
            key=lambda cluster: cluster[0],
            reverse=True,
        )
    )
    return (cluster_representative, cluster_size, cluster_ads)


def fill_sentiment_dict(adjective_counts):
    """Given a dictionary with adjectives and their counts, will compute.

    The sentiment of each of the adjectives using the VADER sentiment analysis package
    and return a dictionary of the adjectives and their sentiments.

    :param adjective_counts: dictionary with adjectives and their counts
    :type adjective_counts: dict
    :return: dictionary, where the keys are the adjectives and the values are tuples of the
        corresponding compound sentiument and binary sentiment
    :rtype: dict
    """
    sentiment_dict = {}
    analyzer = SentimentIntensityAnalyzer()
    for one_topic_adjectives in adjective_counts:
        for adjective, _ in one_topic_adjectives:
            if adjective not in sentiment_dict:
                score = analyzer.polarity_scores(adjective)
                sentiment_dict[adjective] = (
                    score["compound"],
                    (score["pos"] - score["neg"]),
                )
    return sentiment_dict


def sentiments_from_adjectives(adjective_counts, sentiment_dict):
    """Build the weighted average sentiment score from a list of adjetives and their counts.

    :param adjective_counts: list of tuples with adjectives and their counts
    :type adjective_counts: [(str,int)]
    :param sentiment_dict: dictionary with adjectives and their sentiment, as tuple of compound and binary sentiment
    :type sentiment_dict: dict
    :return: a tuple of the weighted averages of compound and binary sentiment
    :rtype: (float,float)
    """
    cumulative_sentimenent_compound = 0
    cumulative_sentimenent_binary = 0
    for adjective, count in adjective_counts:
        cumulative_sentimenent_compound = (
            cumulative_sentimenent_compound + sentiment_dict[adjective][0] * count
        )
        cumulative_sentimenent_binary = (
            cumulative_sentimenent_binary + sentiment_dict[adjective][1] * count
        )

    total = sum(list(zip(*adjective_counts))[1])
    sentimentscore_compound = cumulative_sentimenent_compound / total
    sentimentscore_binary = cumulative_sentimenent_binary / total

    return (sentimentscore_compound, sentimentscore_binary)


def adjective_info(dataframe_topics, dataframe_aspects, vectorizer):
    """Add adjective related information to the dataframes.

    This has two facets:
    -> for each topic cluster similar adjectives, to get a more abstract/readable list
    -> for each topic, use the adjectives to come up with a sentiment classification

    :param dataframe_topics: the dataframe with the topics we want to enrich, needs to have a collum `rawterms`
    :type dataframe_topics: :class:`pandas.DataFrame`
    :param dataframe_aspects: the dataframe with the aspect instances and related adjectives with columsn `aspect` and `descriptor`
    :type dataframe_aspects: :class:`pandas.DataFrame`
    :param vectorizer:  provide embeddings for the adjectives
    :type vectorizer: :class:`_vectorizer.vectorizer`
    :return: the enriched datafromes
    :rtype: (:class:`pandas.DataFrame`,:class:`pandas.DataFrame`)
    """
    # get counts of adjectives connected to a given topic
    dataframe_topics["adjectives"] = dataframe_topics["rawterms"].apply(
        lambda terms: Counter(
            dataframe_aspects[dataframe_aspects["aspect"].isin(terms)]["descriptor"]
        ).most_common()
    )
    dataframe_topics["adjective_clusters"] = dataframe_topics["adjectives"].apply(
        lambda adjective_counts: cluster_adjectives(adjective_counts, vectorizer)
    )

    # sentimentscore for adjectives, based on compund score of Vader sentiment
    # compound refers to mixed score whereas bin refers to binary weight (1/0) of sentiment
    # use one dict, so we don't need to run the classifier more than once for
    # a single adjective
    sentiment_dict = fill_sentiment_dict(dataframe_topics["adjectives"])
    dataframe_topics["sentiment_pair"] = dataframe_topics["adjectives"].apply(
        lambda adjective_counts: sentiments_from_adjectives(
            adjective_counts, sentiment_dict
        )
    )
    dataframe_topics["sentiment_compound"] = dataframe_topics["sentiment_pair"].apply(
        lambda pair: pair[0]
    )
    dataframe_topics["sentiment_binary"] = dataframe_topics["sentiment_pair"].apply(
        lambda pair: pair[1]
    )

    dataframe_aspects["sentiment_compound"] = dataframe_aspects["descriptor"].apply(
        lambda descriptor: sentiment_dict[descriptor][0]
        if descriptor in sentiment_dict
        else 0
    )
    dataframe_aspects["sentiment_binary"] = dataframe_aspects["descriptor"].apply(
        lambda descriptor: sentiment_dict[descriptor][1]
        if descriptor in sentiment_dict
        else 0
    )

    # flip sentiment if adjective is negated
    dataframe_aspects["sentiment_compound"] = dataframe_aspects.apply(
        lambda x: x["sentiment_compound"] * (-1)
        if x["is_negated"]
        else x["sentiment_compound"],
        axis=1,
    )
    dataframe_aspects["sentiment_binary"] = dataframe_aspects.apply(
        lambda x: x["sentiment_binary"] * (-1)
        if x["is_negated"]
        else x["sentiment_binary"],
        axis=1,
    )
    return dataframe_topics, dataframe_aspects



import logging
import subprocess  # nosec
from pathlib import Path
from typing import List

from gensim.models import KeyedVectors
from gensim.test.utils import datapath

from extra_model._errors import ExtraModelError

URL = "http://downloads.cs.stanford.edu/nlp/data/glove.840B.300d.zip"

logger = logging.getLogger(__name__)


def setup(output_path: Path) -> None:
    """Docstring."""
    logging.basicConfig(level="INFO", format="  %(message)s")

    logger.info("")
    logger.info("extra-model-setup will perform the following actions:")
    logger.info("  1. Download embeddings")
    logger.info("  2. Unzip embeddings")
    logger.info("  3. Format embeddings")
    logger.info("This process will take approximately 40 minutes.")
    logger.info("Setup can be safely re-run if exited prematurely.")
    logger.info("")

    if input("  Would you like to continue? [y/n]: ").lower() != "y":
        return

    files_to_cleanup: List[Path] = []

    file_zipped = download_file(URL, output_path)
    files_to_cleanup.append(file_zipped)

    file_unzipped = unzip_file(file_zipped, output_path)
    files_to_cleanup.append(file_unzipped)

    format_file(file_unzipped, output_path)
    cleanup(files_to_cleanup)

    logger.info("Done!")


def download_file(url: str, output_path: Path) -> Path:
    """Download embeddings file."""
    filename = url.split("/")[-1]
    output_file = output_path / filename

    if not output_file.is_file():
        logger.info("Downloading embeddings file.")
        logger.info("This will take approximately 20 minutes.")
        run_subprocess(["curl", url, "-o", str(output_file)])

    else:
        logger.info(f"Embeddings file {output_file} detected, downloading skipped!")

    return output_file


def unzip_file(file: Path, output_path: Path) -> Path:
    """Unzip embeddings file."""
    output_file = output_path / f"{file.stem}.txt"

    if not output_file.is_file():
        logger.info("Unzipping file. This will take approximately 5 minutes.")
        run_subprocess(["unzip", str(file), "-d", str(output_path)])

    else:
        logger.info(f"File {output_file} detected, unzipping skipped!")

    return output_file


def format_file(file: Path, output_path: Path) -> None:
    """Format embeddings file."""
    output_file = output_path / f"{file.stem}"
    if not output_file.is_file():
        logger.info("Formatting file. This will take approximately 10 minutes.")
        glove_file = datapath(str(file))
        model = KeyedVectors.load_word2vec_format(
            glove_file, binary=False, no_header=True
        )
        model.save(str(output_file))

    else:
        logger.info(f"File {output_file} detected, formatting skipped!")


def cleanup(files: List[Path]) -> None:
    """Cleanup setup cruft."""
    logger.info("Cleaning up setup cruft...")

    for file in files:
        logger.info(f"Deleting {file}...")
        file.unlink()


def run_subprocess(command: List[str]) -> None:
    """Run subprocess."""
    try:
        subprocess.run(" ".join(command), check=True, shell=True)  # nosec

    except subprocess.CalledProcessError as e:
        raise ExtraModelError(f"Subprocess error: {e.stderr}")



import pandas as pd

from extra_model._summarize import (
    link_aspects_to_texts,
    link_aspects_to_topics,
    qa,
    set_aspect,
)


def test_summarize__qa():
    # Test is a bit tricky, ideally you'd like to run the complete code to see that the output format
    # matches what is expected by this function. However, that's not really feasible for testing,
    # so we mock the respective dataframes just to see that it's not crashing
    topicframe = pd.DataFrame(
        columns=[
            "topic",
            "rawterms",
            "rawnums",
            "weightedterms",
            "weights",
            "topicID",
            "adjective_clusters",
            "importance",
            "sentiment_compound",
            "sentiment_binary",
        ]
    )
    topicframe.loc[0] = [
        "table",
        ["table"],
        [1],
        ["table"],
        [1],
        1,
        (["small"], [[1]], [[("small", 1)]]),
        1.0,
        1.0,
        1.0,
    ]
    aspectframe = pd.DataFrame(
        [{"aspect": "table", "descriptor": "small", "adcluster": "small", "CiD": 0}]
    )
    textframe = pd.DataFrame([{"Comments": "This is a small table"}])
    # this just needs to not crash
    qa(textframe, aspectframe, topicframe)


def test_summarize__set_aspect():
    topicframe = pd.DataFrame(columns=["rawterms", "topicID", "adjective_clusters"])
    topicframe.loc[0] = [["table"], 1, (["small"], [[1]], [[("small", 1)]])]
    aspectframe = pd.DataFrame(
        [
            {"aspect": "table", "descriptor": "small"},
            {"aspect": "chair", "descriptor": "small"},
        ]
    )
    aspectframe["topicID"] = None
    aspectframe["adcluster"] = None

    topicframe.apply(lambda topic: set_aspect(topic, aspectframe), axis=1)

    assert (
        aspectframe.iloc[0]["topicID"] == 1
        and aspectframe.iloc[0]["adcluster"] == "small"
        and not aspectframe.iloc[1]["topicID"]
        and not aspectframe.iloc[1]["adcluster"]
    )


def test_summarize__link_aspects_to_topics():
    topicframe = pd.DataFrame(columns=["rawterms", "adjective_clusters"])
    topicframe.loc[0] = [["table"], (["small"], [[1]], [[("small", 1)]])]
    aspectframe = pd.DataFrame(
        [
            {"aspect": "table", "descriptor": "small"},
            {"aspect": "chair", "descriptor": "small"},
        ]
    )
    result = link_aspects_to_topics(aspectframe, topicframe)
    assert (
        result.iloc[0]["topicID"] == 0
        and aspectframe.iloc[0]["adcluster"] == "small"
        and not aspectframe.iloc[1]["topicID"]
        and not aspectframe.iloc[1]["adcluster"]
        and topicframe.iloc[0]["topicID"] == 0  # has the topicID been added?
    )


def test_summarize__link_aspects_to_texts():
    aspectframe = pd.DataFrame([{"CiD": 0}, {"CiD": 2}])
    textframe = pd.DataFrame(
        [{"source_guid": 1, "drop": 0}, {"source_guid": 2, "drop": 0}]
    )
    result = link_aspects_to_texts(aspectframe, textframe)
    print(result)
    assert (
        result.iloc[0]["source_guid"] == 1  # join the correct text
        and "CiD" in result.columns.values.tolist()  # keep the right cols
        and "source_guid" in result.columns.values.tolist()  # keep the right cols
        and "drop" not in result.columns.values.tolist()  # drop other columns
    )



import logging
from pathlib import Path

import pytest
from click.testing import CliRunner

from extra_model._cli import entrypoint, entrypoint_setup
from extra_model._errors import ExtraModelError


@pytest.fixture()
def cli_runner():
    return CliRunner()


@pytest.fixture
def run_mock(mocker):
    return mocker.patch("extra_model._cli.run")


@pytest.fixture
def setup_mock(mocker):
    return mocker.patch("extra_model._cli.setup")


INPUT = "/input"
OUTPUT = "/output"
OUTPUT_FILENAME = "result.csv"
EMBEDDINGS_PATH = "/embeddings"


def test_entrypoint__run_raises_no_exception__exit_code_0(cli_runner, run_mock):

    result = cli_runner.invoke(entrypoint, [INPUT], catch_exceptions=False)

    assert result.exit_code == 0


def test_entrypoint__run_raises_ExtraModelError__exit_code_1(cli_runner, run_mock):

    run_mock.side_effect = ExtraModelError

    result = cli_runner.invoke(entrypoint, [INPUT], catch_exceptions=False)

    assert result.exit_code == 1


def test_entrypoint__run_raises_unhandled_exception__exception_raised(
    cli_runner, run_mock
):

    run_mock.side_effect = Exception

    with pytest.raises(Exception):
        cli_runner.invoke(entrypoint, [INPUT], catch_exceptions=False)


def test_entrypoint__debug_set__log_level_set_to_DEBUG(cli_runner, run_mock):

    cli_runner.invoke(entrypoint, [INPUT, "--debug"], catch_exceptions=False)

    assert logging.getLogger("extra_model").level == logging.DEBUG


def test_entrypoint__input_and_output_path_set__arguments_passed_to_run(
    cli_runner, run_mock
):
    cli_runner.invoke(
        entrypoint, [INPUT, OUTPUT, OUTPUT_FILENAME], catch_exceptions=False
    )

    run_mock.assert_called_once_with(
        Path(INPUT), Path(OUTPUT), Path(OUTPUT_FILENAME), Path(EMBEDDINGS_PATH)
    )


def test_entrypoint_setup__setup_raises_no_exception__exit_code_0(
    cli_runner, setup_mock
):

    result = cli_runner.invoke(entrypoint_setup, catch_exceptions=False)

    assert result.exit_code == 0


def test_entrypoint_setup__setup_raises_ExtraModelError__exit_code_1(
    cli_runner, setup_mock
):

    setup_mock.side_effect = ExtraModelError

    result = cli_runner.invoke(entrypoint_setup, catch_exceptions=False)

    assert result.exit_code == 1


def test_entrypoint_setup__setup_raises_unhandled_exception__exception_raised(
    cli_runner, setup_mock
):

    setup_mock.side_effect = Exception

    with pytest.raises(Exception):
        cli_runner.invoke(entrypoint_setup, catch_exceptions=False)


def test_entrypoint_setup__output_path_set__argument_passed_to_setup(
    cli_runner, setup_mock
):

    cli_runner.invoke(entrypoint_setup, [OUTPUT], catch_exceptions=False)

    setup_mock.assert_called_once_with(Path(OUTPUT))



import logging
from pathlib import Path

import pandas as pd

from extra_model._models import ExtraModel

MODELS_FOLDER = Path("./embeddings")
OUTPUT_FILE = Path("result.csv")

logger = logging.getLogger(__name__)


def run(
    input_path: Path,
    output_path: Path,
    output_filename: Path = OUTPUT_FILE,
    embeddings_path: Path = MODELS_FOLDER,
) -> None:
    """Docstring."""
    logging.basicConfig(format="  %(message)s")

    extra_model = ExtraModel(models_folder=embeddings_path)
    extra_model.load_from_files()

    logger.info(f"Loading data from {input_path}")
    input_data = pd.read_csv(input_path)

    logger.info("Running `extra-model`")
    results_raw = extra_model.predict(comments=input_data.to_dict("records"))
    results = pd.DataFrame(results_raw)

    if not output_path.exists():
        logger.info(f"Creating folder {output_path}")
        output_path.mkdir(parents=True)

    logger.info(f"Saving output to {output_path / output_filename}")
    results.to_csv(output_path / output_filename, encoding="utf-8", index=False)



import os

import pandas
import pytest
from gensim.models import KeyedVectors

from extra_model._adjectives import (
    adjective_info,
    cluster_adjectives,
    fill_sentiment_dict,
    sentiments_from_adjectives,
)
from extra_model._vectorizer import Vectorizer


@pytest.fixture()
def vec():
    # preprocess plain-text test embeddings to proper binary format for vectorizer
    glove_file = "tests/resources/test_adjectives.vec"
    prepro_file = "tests/resources/test_adjectives.prepro"
    model = KeyedVectors.load_word2vec_format(glove_file, binary=False, no_header=True)
    model.save(prepro_file)

    yield Vectorizer(prepro_file)

    # cleanup
    os.remove(prepro_file)


def test__cluster_adjectives__empty(vec):
    # should return False if only unknown adjectives appear
    assert not cluster_adjectives(adjective_counts=[("asdf", 1)], vectorizer=vec)


def test__cluster_adjectives__single(vec):
    # test special case if there is only one adjective
    assert cluster_adjectives(adjective_counts=[("small", 1)], vectorizer=vec) == (
        ["small"],
        [[1]],
        [[("small", 1)]],
    )


def test__cluster_adjectives__common(vec):
    # test the common case with more than one cluster
    # test embeddings for 'ugly' and 'beautiful' are artificially close together to test
    # antonym removal functionality
    adjective_counts = [("small", 23), ("little", 22), ("Beautiful", 19), ("Ugly", 1)]
    assert cluster_adjectives(adjective_counts=adjective_counts, vectorizer=vec) == (
        ("small", "Beautiful", "Ugly"),
        (45, 19, 1),
        ([("small", 23), ("little", 22)], [("Beautiful", 19)], [("Ugly", 1)]),
    )


def test__fill_sentiment_dict():
    assert fill_sentiment_dict([[("Ugly", 1)]]) == {"Ugly": (-0.5106, -1.0)}


def test__sentiments_from_adjectives():
    assert sentiments_from_adjectives(
        [("pretty", 3), ("ugly", 1)], {"pretty": (0.5, 1), "ugly": (-0.5, -1)}
    ) == (0.25, 0.5)


def test__adjective_info(vec):
    # important part here is to test the inversion of the sentiment of negated adjectives
    df_topics = pandas.DataFrame([{"rawterms": ["chair"]}])
    df_aspects = pandas.DataFrame(
        [
            {"aspect": "chair", "descriptor": "beautiful", "is_negated": False},
            # should not drag down score
            {"aspect": "chair", "descriptor": "ugly", "is_negated": True},
            # should  drag down score
            {"aspect": "chair", "descriptor": "ugly", "is_negated": False},
            {"aspect": "chair", "descriptor": "small", "is_negated": False},
            {"aspect": "chair", "descriptor": "little", "is_negated": False},
            # test for unknown adjectives
            {"aspect": "chair", "descriptor": "asdf", "is_negated": False},
        ]
    )
    dft, dfa = adjective_info(df_topics, df_aspects, vec)
    assert (
        # beautiful has potisitve sentiment
        dfa.iloc[0]["sentiment_compound"] > 0
        and dfa.iloc[0]["sentiment_binary"] == 1
        # negated ugly has positive sentiment
        and dfa.iloc[1]["sentiment_compound"] > 0
        and dfa.iloc[1]["sentiment_binary"] == 1
        # straight ugly has negative sentiment
        and dfa.iloc[2]["sentiment_compound"] < 0
        and dfa.iloc[2]["sentiment_binary"] == -1
        # small has neutral sentiment
        and dfa.iloc[3]["sentiment_compound"] == 0
        and dfa.iloc[3]["sentiment_binary"] == 0
        # unkown word has neutral sentiment
        and dfa.iloc[5]["sentiment_compound"] == 0
        and dfa.iloc[5]["sentiment_binary"] == 0
        # do we have correct counts?
        and dft.iloc[0]["adjectives"]
        == [("ugly", 2), ("beautiful", 1), ("small", 1), ("little", 1), ("asdf", 1)]
        and dft.iloc[0]["adjective_clusters"]
        # clustering drops the unkonw word
        == (
            ("ugly", "beautiful", "small"),
            (3, 1, 1),
            ([("ugly", 2), ("little", 1)], [("beautiful", 1)], [("small", 1)]),
        )
        # negated and bare 'ugly' should cancel,
        # TODO: reactivate this test, once adjectives has been reviewed
        # https://projecthub.service.csnzoo.com/browse/AUXEM-42
        # and dft.iloc[0]["sentiment_compound"] == pytest.approx(0)
        # and dft.iloc[0]["sentiment_binary"] == pytest.approx(0)
    )



import pytest

from extra_model._run import run


@pytest.fixture
def extra_model_mock(mocker):
    return mocker.patch("extra_model._run.ExtraModel")


@pytest.fixture
def pandas_mock(mocker):
    return mocker.patch("extra_model._run.pd")


def output_path(mocker):
    return mocker.Mock()


def test_run__csv_created(mocker, extra_model_mock, pandas_mock):

    run(mocker.MagicMock(), mocker.MagicMock())

    assert extra_model_mock.called
    assert pandas_mock.DataFrame.called
    assert pandas_mock.DataFrame.return_value.to_csv.called


def test_run__output_path_exists__output_path_not_created(
    mocker, extra_model_mock, pandas_mock
):

    output_path = mocker.MagicMock()
    output_path.exists.return_value = True

    run(mocker.MagicMock(), output_path)

    assert output_path.exists.called
    assert not output_path.mkdir.called


def test_run__output_path_does_not_exist__output_path_created(
    mocker, extra_model_mock, pandas_mock
):

    output_path = mocker.MagicMock()
    output_path.exists.return_value = False

    run(mocker.MagicMock(), output_path)

    assert output_path.exists.called
    assert output_path.mkdir.called



import logging

import cld3
import pandas as pd

logger = logging.getLogger(__name__)


def filter(dataframe):
    """Filter a dataframe for language and text length.

    The following rules apply:
    1. Only comments with at least 20 characters retained.
    2. Only comments in English are retained.
    3. All unprintable unicode characters are removed.

    :param dataframe: dataframe to be filtered. Must have column "Comments"
    :type pd.DataFrame
    :return: Filtered dataframe
    :rtype: pd.DataFrame
    """
    # 'None' comments would lead to cashes later
    dataframe = dataframe[pd.notnull(dataframe["Comments"])]
    # reset here to avoid "copy of slice" warning
    dataframe = dataframe.reset_index(drop=True)

    # filter on comment length
    dataframe.loc[:, "cl"] = dataframe.Comments.apply(
        lambda com: len(com) if com is not None else 0
    )
    dataframe = dataframe[dataframe.cl > 20]

    # remove problematic unicode characters
    dataframe.loc[:, "Comments"] = dataframe.Comments.apply(
        lambda com: "".join(x for x in com if x.isprintable())
    )

    # detect language and filter english. If it's 'unknown' it's probably
    # still english
    dataframe.loc[:, "lang"] = dataframe.Comments.apply(
        lambda com: cld3.get_language(com).language
    )
    dataframe = dataframe[(dataframe["lang"] == "en") | (dataframe["lang"] == "un")]

    # drop auxiliary columns again, re-index
    dataframe.drop(["cl", "lang"], axis="columns", inplace=True)
    dataframe.reset_index(inplace=True, drop=True)

    return dataframe



import json
import os

import pandas as pd
import pytest

from extra_model._models import ExtraModelBase, ModelBase, extra_factory

ExtraModel = extra_factory()


class MyCustomBase:
    pass


def test_extra_factory__no_custom_bases_passed__class_has_ModelBase_as_parent():

    Model = extra_factory()

    assert issubclass(Model, ModelBase)


def test_extra_factory__no_custom_bases_passed__class_has_ExtraModelBase_as_parent():

    Model = extra_factory()

    assert issubclass(Model, ExtraModelBase)


def test_extra_factory__custom_bases_passed__class_has_custom_base_as_parent():

    Model = extra_factory(MyCustomBase)

    assert issubclass(Model, MyCustomBase)


def test_extra_factory__multiple_custom_bases_passed__class_has_all_custom_bases_as_parents():
    class MyOtherCustomBase:
        pass

    Model = extra_factory((MyCustomBase, MyOtherCustomBase))

    assert issubclass(Model, MyCustomBase)
    assert issubclass(Model, MyOtherCustomBase)


def test_extra_factory__custom_bases_passed__class_does_not_have_ModelBase_as_parent():

    Model = extra_factory(MyCustomBase)

    assert not issubclass(Model, ModelBase)


def test_extra_factory__custom_bases_passed__class_has_extraModelBase_as_parent():

    Model = extra_factory(MyCustomBase)

    assert issubclass(Model, ExtraModelBase)


@pytest.fixture
def tmp_untrained_res_models_folder_ExtraModel():
    resources_folder = os.path.dirname(os.path.realpath(__file__)) + "/resources/"
    untrained = ExtraModel(
        models_folder=resources_folder, embedding_type="small_embeddings"
    )
    return untrained


@pytest.fixture
def tmp_untrained_tmp_models_folder_ExtraModel(tmpdir_factory):
    tmp_folder = tmpdir_factory.mktemp("tmp_models")
    untrained = ExtraModel(models_folder=tmp_folder, embedding_type="small_embeddings")
    yield untrained


@pytest.fixture
def tmp_trained_ExtraModel():
    resources_folder = os.path.dirname(os.path.realpath(__file__)) + "/resources/"
    trained = ExtraModel(
        models_folder=resources_folder, embedding_type="small_embeddings"
    )
    trained.load_from_files()
    return trained


@pytest.fixture
def test_comments():
    input_ = pd.read_csv("./tests/resources/100_comments.csv")
    return input_.to_dict("records")


def test_create(tmp_untrained_tmp_models_folder_ExtraModel, test_comments):
    # test for initialization
    assert not tmp_untrained_tmp_models_folder_ExtraModel.is_trained
    # test that prediction fails without loading
    with pytest.raises(RuntimeError):
        tmp_untrained_tmp_models_folder_ExtraModel.predict(comments=test_comments)
    # for filekey in tmp_untrained_tmp_models_folder_ExtraModel.filenames:
    #     assert filekey in tmp_untrained_tmp_models_folder_ExtraModel.storage_metadata()


def test_load_from_files(tmp_untrained_res_models_folder_ExtraModel):
    assert not tmp_untrained_res_models_folder_ExtraModel.is_trained
    tmp_untrained_res_models_folder_ExtraModel.load_from_files()
    assert tmp_untrained_res_models_folder_ExtraModel.is_trained


def test_predict(tmp_trained_ExtraModel, test_comments):
    # Extra is an unsupervised algorithm, so not possible to guarantee certain output
    res = tmp_trained_ExtraModel.predict(comments=test_comments)
    res_names = set(pd.DataFrame(res).columns.values)
    required_names = set(tmp_trained_ExtraModel.api_spec_names.values())
    assert len(required_names.difference(res_names)) == len(
        res_names.difference(required_names)
    )
    try:
        json.dumps(res)
    except Exception:
        pytest.fail("Serialization failed")

    # extra should fail with correct error for this input
    test_comments_1 = [
        {"CommentId": 1, "Comments": "Hey ho"},
        {"CommentId": 2, "Comments": "comment number two"},
    ]
    with pytest.raises(ValueError):
        tmp_trained_ExtraModel.predict(comments=test_comments_1)

    # extra should succeed with this input
    test_comments_2 = [
        {
            "CommentId": 986,
            "Comments": "The problem I have is that they charge $11.99 for a sandwich that is no bigger than a Subway sub (which offers better and more amount of vegetables).",
        },
        {"CommentId": 860, "Comments": "Worst food/service I've had in a while"},
    ]
    res = tmp_trained_ExtraModel.predict(comments=test_comments_2)
    assert len(res) > 0
